# Generated by Django 2.2.23 on 2022-03-01 08:52

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("nuntius", "0022_add_start_and_end_dates_to_campaigns")]

    operations = [
        migrations.AlterField(
            model_name="pushcampaign",
            name="notification_tag",
            field=models.CharField(
                blank=True,
                help_text="Tagging a notification allows replacing it by pushing another with the same tag value",
                max_length=255,
                null=True,
                verbose_name="Notification tag",
            ),
        )
    ]
